import math
from abc import ABC
from .joingraph import JoinGraph
from .semiring import *
from .aggregator import Aggregator, Annotation, Message
from .cjt import CJT
from queue import PriorityQueue


class App(ABC):
    def __init__(self):
        pass
    
class DummyModel(App):
    def __init__(self):
        super().__init__()
        self.semi_ring = varSemiRing()
        self.prefix = "joinboost_tmp_"
    
    def fit(self,
           jg: JoinGraph):
        jg._preprocess()
        self.jg = jg

        # compute the total average
        agg_exp = self.semi_ring.col_sum(s=self.jg.get_target_var(), c = '1')
        TS, TC = self.jg.exe.execute_spja_query(agg_exp,
                                              [self.jg.get_target_relation()],
                                              mode = 3)[0]
        mean = TS / TC
        self.semi_ring.set_semi_ring(TS, TC)
        
        self.count_ = TC
        self.constant_ = mean

    def predict(self):
        return self.constant_
    

class DecisionTree(DummyModel):
    def __init__(self,                
                 max_leaves: int = 31,
                 learning_rate: float = 1, 
                 max_depth: int = 6):
        super().__init__()
        self.max_leaves = max_leaves
        self.learning_rate = learning_rate
        self.max_depth = max_depth
        
    def fit(self,
           jg: JoinGraph):
        self.model_def = []
        super().fit(jg)
        self.cjt = CJT(semi_ring=self.semi_ring, join_graph=self.jg)
        self.cjt.lift(jg.get_target_var() + "- (" + str(self.constant_) + ")")
        self.semi_ring.set_semi_ring(0, self.count_)
        self.train_one()
    
    def train_one(self, last = True):
        # store (node_id) -> cjt
        self.nodes = {}
        self.nodes[0] = self.cjt
        
        # store a pq of split_candidates, sorted on the criteria
        # this one is a bit complex. TODO: simplify or make it a class
        self.split_candidates = PriorityQueue()
        
        # leaf_nodes is used to compute the final models
        self.leaf_nodes = []
        
        self._build_tree()
        if last:
            self._update_error()
            
        self._build_model()
        self._clean_messages()
        
    def _build_model(self):
        cur_model_def = []
        for cur_cjt in self.leaf_nodes:
            annotations = cur_cjt.get_all_parsed_annotations()
            TS, TC = cur_cjt.get_semi_ring().get_value()
            pred = TS / TC * self.learning_rate
            if annotations:
                cur_model_def.append((pred, annotations))
        if cur_model_def:
            self.model_def.append(cur_model_def)
        
    def compute_rmse(self, test_table: str):
        # TODO: refactor
        view = self.cjt.exe.case_query(test_table, '+', 'prediction', str(self.constant_),
                                       self.model_def, [self.cjt.get_target_var()])
        predict_agg = {'RMSE': ('SQRT(AVG(POW(' + self.cjt.get_target_var() + ' - prediction,2)))',
                                Aggregator.IDENTITY)}
        predict = self.cjt.exe.execute_spja_query(predict_agg, [view])
        return self.cjt.exe.select_all(predict)[0]

    def _clean_messages(self):
        for cjt in self.nodes.values():
            cjt.clean_message()

    def _comp_annotations(self, r_name: str, attr: str, cur_value: str, obj: float, expanding_cjt: CJT):
        attr_type = expanding_cjt.get_relation_schema()[r_name][attr]
        # TODO: remove window_query and everything is spja
        if attr_type == 'LCAT':
            group_by = [attr]
            absoprtion_view = expanding_cjt.absorption(r_name, [attr])
            agg_exp = {attr: (attr, Aggregator.IDENTITY),
                       'object': (('s', 'c'), Aggregator.DIV),
                       's': ('s', Aggregator.IDENTITY),
                       'c': ('c', Aggregator.IDENTITY)}
            obj_view = self.cjt.exe.execute_spja_query(agg_exp, [absoprtion_view])
            view_ord_by_obj = self.cjt.exe.window_query(obj_view, [attr], 'object', ['s', 'c'])
            attr_view = self.cjt.exe.execute_spja_query({attr: (attr, Aggregator.IDENTITY)},
                                                        [view_ord_by_obj],
                                                        annotations=['s/c <=' + str(obj)])
            attrs = [str(x[0])  for x in self.cjt.exe.select_all(attr_view)]
            l_annotation = (attr, Annotation.IN, attrs)
            r_annotation = (attr, Annotation.NOT_IN, attrs)
        elif cur_value == 'NULL':
            l_annotation = (attr, Annotation.NULL, Annotation.NULL)
            r_annotation = (attr, Annotation.NOT_NULL, Annotation.NOT_NULL)
        elif attr_type == 'NUM':
            l_annotation = (attr, Annotation.NOT_GREATER, cur_value)
            r_annotation = (attr, Annotation.GREATER, cur_value)
        elif attr_type == 'CAT':
            l_annotation = (attr, Annotation.NOT_DISTINCT, cur_value)
            r_annotation = (attr, Annotation.DISTINCT, cur_value)
        else:
            raise Exception('Unsupported Split')
        return l_annotation, r_annotation

    # get best split of current cjt
    def _get_best_split(self, cjt_id: int, cjt_depth: int):
        cjt = self.nodes[cjt_id]
        cur_semi_ring = cjt.get_semi_ring()
        attr_meta = self.cjt.get_relation_schema()
        
        # criteria, (relation name, split attribute, split value, new s, new c)
        best_criteria, best_criteria_ann = 0, ('', '', 0, 0, 0)
        
        if cjt_depth == self.max_depth:
            self.split_candidates.put((-best_criteria, cjt_depth,) + best_criteria_ann + (cjt_id,))
            return
        
        for r_name in cjt.get_relations():
            for attr in cjt.get_relation_features(r_name):
                attr_type, group_by = self.cjt.get_type(r_name, attr), [attr]
                absoprtion_view = cjt.absorption(r_name, group_by, mode=4)
                ts, tc = cur_semi_ring.get_value()
                if attr_type == 'NUM':
                    # TODO: make ['c', 's'] be something we can get from semi-ring, for different metrics
                    agg_exp = cur_semi_ring.col_sum()
                    agg_exp[attr] = (attr, Aggregator.IDENTITY)
                    view_to_max = self.cjt.exe.execute_spja_query(agg_exp,
                                                                  [absoprtion_view], 
                                                                  window_by=[attr],
                                                                  mode=4)
                
                # TODO: remove window function
                elif attr_type == 'LCAT':
                    agg_exp = {attr: (attr, Aggregator.IDENTITY),
                               'object': (('s', 'c'), Aggregator.DIV),
                               's': ('s', Aggregator.IDENTITY),
                               'c': ('c', Aggregator.IDENTITY)}
                    obj_view = self.cjt.exe.execute_spja_query(agg_exp, [absoprtion_view])
                    view_to_max = self.cjt.exe.window_query(obj_view, [attr], 'object', ['s', 'c'])
                else:
                    view_to_max = absoprtion_view
                
                # TODO: move this logic somewhere else
                l2_agg_exp = {
                    attr: (attr, Aggregator.IDENTITY),
                    'criteria': ('CASE WHEN ' + str(tc) + ' > c THEN (-(CAST(' + str(ts) + ' AS DOUBLE)/' + str(tc) +
                                ')* ' + str(ts) + ' + (s/c)*s + (' + str(ts) + '-s)/(' + str(tc) + '-c)*(' + str(ts) +
                                '-s)) ELSE 0 END', Aggregator.IDENTITY),
                    'c': ('c', Aggregator.IDENTITY),
                    's': ('s', Aggregator.IDENTITY),
                }
                view_max = self.cjt.exe.execute_spja_query(l2_agg_exp, 
                                                          [view_to_max], 
                                                          order_by='criteria DESC', 
                                                          limit=1, 
                                                          mode=4)
                
                results = self.cjt.exe.select_all(view_max)
                if not results:
                    continue
                cur_value, cur_criteria, c, s = results[0]
                # print(results[0])
                if cur_criteria >= best_criteria:
                    best_criteria = cur_criteria
                    # relation name, split attribute, split value, new s, new c  
                    best_criteria_ann = (r_name, attr, str(cur_value), s, c)
        self.split_candidates.put((-best_criteria, cjt_depth,) + best_criteria_ann + (cjt_id,))
        
    # split the semi-ring according to current split
    def split_semi_ring(self, 
                        total_semi_ring: varSemiRing, 
                        left_semi_ring: varSemiRing):
        return left_semi_ring, total_semi_ring - left_semi_ring
    
    # don't update error for single ecision tree 
    def _update_error(self):
        pass
    
    def _get_split_cjt(self, expanding_cjt: CJT, l_semi_ring: varSemiRing, r_semi_ring: varSemiRing):
        l_cjt, r_cjt = expanding_cjt.copy_cjt(l_semi_ring), expanding_cjt.copy_cjt(r_semi_ring)
        next_id = len(self.nodes)
        
        self.nodes[next_id] = l_cjt
        self.nodes[next_id + 1] = r_cjt
        
        return l_cjt, r_cjt, next_id, next_id + 1

    def _build_tree(self):
        self.cjt.calibration()
        self._get_best_split(0, 0)
        
        # while there are beneficial splits and doesn't read max leaves 
        while not self.split_candidates.empty() and \
              self.split_candidates.queue[0][0] < 0 and \
              self.split_candidates.qsize() < self.max_leaves:
            criteria, cur_level, r_name, attr, cur_value, s, c, c_id = self.split_candidates.get()
            expanding_cjt = self.nodes[c_id]
            
            l_semi_ring, r_semi_ring = self.split_semi_ring(expanding_cjt.get_semi_ring(), varSemiRing(s, c))
            l_cjt, r_cjt, l_id, r_id = self._get_split_cjt(expanding_cjt=expanding_cjt,
                                                           l_semi_ring=l_semi_ring,
                                                           r_semi_ring=r_semi_ring)
            # TODO: objective has some rounding problem
            # currently it has an ugly solution. find a better solution
            l_annotations, r_annotations = self._comp_annotations(r_name=r_name, 
                                                                  attr=attr,
                                                                  cur_value=cur_value,
                                                                  obj=math.ceil(s / c * 100) / 100,
                                                                  expanding_cjt=expanding_cjt)
            
            # add annotations according to split conditions
            l_cjt.add_annotations(r_name, l_annotations)
            r_cjt.add_annotations(r_name, r_annotations)
            
            # for the leaf split_candidates that can't be splitted (e.g. meet max depth)
            # we still need message passing to fact table for semi-join selection
            # but not necessarily downward_message_passing. 
            # Can be optimized to upward_message_passing(fact)
            l_cjt.downward_message_passing(r_name)
            r_cjt.downward_message_passing(r_name)
            
            self._get_best_split(l_id, cur_level + 1)
            self._get_best_split(r_id, cur_level + 1)
                
        self.leaf_nodes = [self.nodes[ele[-1]] for ele in self.split_candidates.queue]


        
class GradientBoosting(DecisionTree):
    # TODO: add some checks. E.g., paramters have to be positive
    def __init__(self,
                 max_leaves: int = 31,
                 learning_rate: float = 1, 
                 max_depth: int = 6,
                 iteration: int = 1):
        super(GradientBoosting, self).__init__(max_leaves,learning_rate,max_depth)
        self.iteration = iteration
    
    def fit(self,
           jg: JoinGraph):
        total_update_time = 0
        super().fit(jg)
        
        for _ in range(self.iteration - 1):
            self.train_one()

    def _update_error(self):
        for cur_cjt in self.leaf_nodes:
            cur_cond = []
            TS, TC = cur_cjt.get_semi_ring().get_value()
            pred = TS / TC * self.learning_rate
            _, join_conds = cur_cjt._get_income_messages(cur_cjt.get_target_relation(), condition=2)
            self.cjt.exe.update_query("s=s-(" + str(pred) + ")",
                                     cur_cjt.get_target_relation(),
                                     join_conds)



from abc import ABC, abstractmethod
from .aggregator import *
import time

class Executor(ABC):
    '''Assume input data are csvs'''

    def __init__(self):
        self.view_id = 0

    def get_next_name(self):
        name = 'joinboost_tmp_' + str(self.view_id)
        self.view_id += 1
        return name
    
    def get_schema(self, table):
        pass
    
    def select_all(self, table):
        pass

    def delete_table(self, table: str):
        pass

    def load_table(self, table_name: str, data_dir: str):
        pass

    def case_query(self, from_table: str, operator: str, cond_attr: str, base_val: str,
                   case_definitions: list, select_attrs: list = [], table_name: str = None):
        pass

    def window_query(self, view: str, select_attrs: list, base_attr: str, cumulative_attrs: list):
        pass

    def execute_spja_query(self,
                           mode: int,
                           aggregate_expressions: dict,
                           in_msgs: list = [],
                           from_table: str = '',
                           group_by: list = [],
                           where_conds: dict = {},
                           annotations: list = [],
                           left_join: dict = {},
                           table_name: str = '',
                           replace: bool = True) -> str:
        pass


class DuckdbExecutor(Executor):
    def __init__(self, conn, debug=False):
        super().__init__()
        self.conn = conn
        self.debug = debug
        
    def get_schema(self, table):
        # duckdb stores table info in [cid, name, type, notnull, dflt_value, pk]
        table_info = self._execute_query('PRAGMA table_info(' + table + ')')
        return [x[1] for x in table_info]
        
    def _gen_sql_case(self, leaf_conds: list):
        conds = []
        for leaf_cond in leaf_conds:
            cond = 'CASE\n'
            for (pred, annotations) in leaf_cond:
                cond += ' WHEN ' + ' AND '.join(annotations) + \
                        ' THEN CAST(' + str(pred) + ' AS DOUBLE)\n'
            cond += 'ELSE 0 END\n'
            conds.append(cond)
        return conds

    def select_all(self, table: str):
        return self.execute_spja_query({None: ('*', Aggregator.IDENTITY)}, 
                                       from_tables=[table], 
                                       mode=3)
    
    def delete_table(self, table: str):
        sql = 'DROP TABLE IF EXISTS ' + table + ';\n'
        self._execute_query(sql)
        
    # TODO: remove it
    def window_query(self, view: str, select_attrs: list, base_attr: str, cumulative_attrs: list):
        view_name = self.get_next_name()
        sql = 'CREATE OR REPLACE VIEW ' + view_name + ' AS SELECT * FROM\n'
        sql += '(\nSELECT ' + ",".join(select_attrs)
        for attr in cumulative_attrs:
            sql += ',SUM(' + attr + ') OVER joinboost_window as ' + attr
        sql += '\nFROM ' + view
        sql += ' WINDOW joinboost_window AS (ORDER BY ' + base_attr + ')\n)'
        self._execute_query(sql)
        return view_name

    # {case: value} operator {case: value} ...
    def case_query(self, from_table: str, operator: str, cond_attr: str, base_val: str,
                   case_definitions: list, select_attrs: list = [], table_name: str = None):
        # print(conditions)
        if not select_attrs:
            attrs = self._execute_query('PRAGMA table_info(' + from_table + ')')
            for attr in attrs:
                if attr != cond_attr: select_attrs.append(attr[1])
        if not table_name:
            view = self.get_next_name()
        else:
            view = table_name
        sql = 'CREATE OR REPLACE TABLE ' + view + ' AS\n'
        sql += 'SELECT ' + ','.join(select_attrs) + ','
        sql += base_val
        for case_definition in case_definitions:
            sql += operator + '\nCASE\n'
            for val, cond in case_definition:
                sql += ' WHEN ' + ' AND '.join(cond) + ' THEN CAST(' + str(val) + ' AS DOUBLE)\n'
            sql += 'ELSE 0 END\n'
        sql += 'AS ' + cond_attr + ' FROM ' + from_table
        self._execute_query(sql)
        return view
    
    def update_query(self,
                     update_expression,
                     table,
                     select_conds: list = []):
        sql = "UPDATE " + table + " SET " + update_expression + " \n"
        if len(select_conds) > 0:
            sql += "WHERE " + " AND ".join(select_conds) + "\n"
        self._execute_query(sql)
    
    # mode = 1 will create the query result as table and return table name
    # mode = 2 will create the query result as table and return view name
    # mode = 3 will create the query result and return the result
    # mode = 4 will create the sql query and return the query (for nested query)
    def execute_spja_query(self, 
                           aggregate_expressions: dict = {},
                           from_tables: list = [],
                           select_conds: list = [],
                           group_by: list = [], 
                           window_by: list = [],
                           table_name: str = None,
                           order_by: str = None,
                           limit: int = None,
                           replace: bool = True,
                           mode: int = 5):
        
        spja = self.spja_query(aggregate_expressions=aggregate_expressions,
                               from_tables=from_tables,
                               select_conds = select_conds,
                               window_by=window_by,
                               group_by=group_by, 
                               order_by=order_by,
                               limit=limit,)
        
        if mode == 1:
            name_ = (table_name if table_name is not None else self.get_next_name())
            entity_type_ = 'TABLE '
            sql = 'CREATE ' + ('OR REPLACE ' if replace else '') + entity_type_  + name_ + ' AS '
            sql += spja
            self._execute_query(sql)
            return name_
        elif mode == 3:
            return self._execute_query(spja)
        elif mode == 4:
            sql = '(' + spja + ')'
            return sql    
        else:
            name_ = (table_name if table_name is not None else self.get_next_name())
            entity_type_ = 'TABLE '
            sql = 'CREATE ' + ('OR REPLACE ' if replace else '') + entity_type_  + name_ + ' AS '
            sql += spja
            self._execute_query(sql)
            return name_
        
    
    
    def spja_query(self, 
                   aggregate_expressions: dict,
                   from_tables: list = [],
                   select_conds: list = [],
                   window_by: list = [],
                   group_by: list = [], 
                   order_by: str = None,
                   limit: int = None,
                   ):
        
        parsed_aggregate_expressions = []
        for target_col, aggregation_spec in aggregate_expressions.items():
            para, agg = aggregation_spec
            parsed_aggregate_expressions.append(parse_agg(agg, para) \
                                + (' OVER joinboost_window ' if len(window_by) > 0 and is_agg(agg) else '')\
                                + (' AS ' + target_col if target_col is not None else ''))
                                                
        
        sql = 'SELECT ' + ', '.join(parsed_aggregate_expressions) + '\n'
        sql += "FROM " + ",".join(from_tables) + '\n'
        
        if len(select_conds) > 0:
            sql += "WHERE " + " AND ".join(select_conds) + '\n'
        if len(window_by) > 0:
            sql += 'WINDOW joinboost_window AS (ORDER BY ' + ','.join(window_by) + ')\n'
        if len(group_by) > 0:
            sql += "GROUP BY " + ",".join(group_by) + '\n'
        if order_by is not None:
            sql += 'ORDER BY ' + order_by + '\n'
        if limit  is not None:
            sql += 'LIMIT ' + str(limit) + '\n'
        return sql

    def _execute_query(self, q):
        start_time = time.time()
        if self.debug:
            print(q)
        self.conn.execute(q)
        elapsed_time = time.time() - start_time
        
        if self.debug:
            print(elapsed_time)
        
        result = None
        try:
            result = self.conn.fetchall()
        except Exception as e:
            print(e)
        return result




from enum import Enum

# TODO: make aggregator class (like operator in database), so we can do simple composition and optimization
Aggregator = Enum('Aggregator', 'SUM MAX MIN SUB SUM_PROD PROD DIV COUNT DISTINCT_COUNT IDENTITY')
Annotation = Enum('NULL', 'NULL NOT_NULL NOT_GREATER GREATER DISTINCT NOT_DISTINCT IN NOT_IN')
Message = Enum('Message', 'IDENTITY SELECTION FULL UNDECIDED')

def parse_agg(agg, para):
    if agg == Aggregator.SUM:
        assert isinstance(para, str)
        return 'SUM(' + para + ')'
    elif agg == Aggregator.SUM_PROD:
        assert isinstance(para, dict)
        _tmp = [key + '.' + value for key, value in para.items()]
        return 'SUM(' + '*'.join(_tmp) + ')'
    elif agg == Aggregator.DISTINCT_COUNT:
        assert isinstance(para, str)
        return 'COUNT(DISTINCT(' + para + '))'
    elif agg == Aggregator.COUNT:
        assert isinstance(para, str)
        return 'COUNT(' + para + ')'
    elif agg == Aggregator.IDENTITY:
        return str(para)
    elif agg == Aggregator.PROD:
        assert isinstance(para, list)
        _tmp = ['CAST(' + val + ' AS DOUBLE)' for val in para]
        return '*'.join(_tmp) 
    elif agg == Aggregator.SUB:
        assert isinstance(para, tuple)
        return str(para[0]) + ' - ' + str(para[1])
    elif agg == Aggregator.DIV:
        assert isinstance(para, tuple)
        return str(para[0]) + ' / ' + str(para[1])
    else:
        raise Exception('Unsupported Semiring Expression')
        
# temp way to check if agg is aggregation or not
def is_agg(agg):
    if agg == Aggregator.SUM:
        return True
    elif agg == Aggregator.SUM_PROD:
        return True
    elif agg == Aggregator.DISTINCT_COUNT:
        return True
    elif agg == Aggregator.COUNT:
        return True
    elif agg == Aggregator.PROD:
        return True
    return False

def parse_ann(annotations: dict, prepend_relation=False):
    select_conds = []
    for r_name, annotations in annotations.items():
        for ann in annotations:
            if prepend_relation:
                attr = ann[0]
            else:
                attr = r_name + '.' + ann[0]
            if ann[1] == Annotation.IN:
                assert isinstance(ann[2], list)
                _tmp = ["'" + str(ele) + "'" for ele in ann[2]]
                select_conds.append(attr + ' in (' + ','.join(_tmp) + ')')
            elif ann[1] == Annotation.NOT_IN:
                assert isinstance(ann[2], list)
                _tmp = ["'" + str(ele) + "'" for ele in ann[2]]
                select_conds.append(attr + ' not in (' + ','.join(_tmp) + ')')
            elif ann[1] == Annotation.NOT_DISTINCT:
                select_conds.append(attr + " == '" + str(ann[2]) + "'")
            elif ann[1] == Annotation.DISTINCT:
                select_conds.append(attr + " != '" + str(ann[2]) + "'")
            elif ann[1] == Annotation.NOT_GREATER:
                select_conds.append(attr + ' <= ' + str(ann[2]))
            elif ann[1] == Annotation.GREATER:
                select_conds.append(attr + ' > ' + str(ann[2]))
            elif ann[1] == Annotation.NULL:
                select_conds.append(attr + ' != ' + attr)
            elif ann[1] == Annotation.NOT_NULL:
                select_conds.append(attr + ' == ' + attr)
    return select_conds


import unittest
import math
import pandas as pd
import duckdb
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.metrics import mean_squared_error
from joinboost.executor import DuckdbExecutor
from joinboost.joingraph import JoinGraph
from joinboost.app import DecisionTree,GradientBoosting

class TestDecision(unittest.TestCase):
        
    def test_synthetic(self):
        join = pd.read_csv("../data/synthetic/RST.csv")
        con = duckdb.connect(database=':memory:')
        con.execute("CREATE TABLE R AS SELECT * FROM '../data/synthetic/R.csv'")
        con.execute("CREATE TABLE S AS SELECT * FROM '../data/synthetic/S.csv'")
        con.execute("CREATE TABLE T AS SELECT * FROM '../data/synthetic/T.csv'")
        con.execute("CREATE TABLE test AS SELECT * FROM '../data/synthetic/RST.csv'")
        x = ["A", "B", "D", "E", "F"]
        y = "C"

        exe = DuckdbExecutor(con, debug=False)
    
        dataset = JoinGraph(exe=exe)
        dataset.add_relation('R', ['B', 'D'], y = 'C')
        dataset.add_relation('S', ['A', 'E'])
        dataset.add_relation('T', ['F'])
        dataset.add_join("R", "S", ["A"], ["A"])
        dataset.add_join("R", "T", ["B"], ["B"])
        
        depth = 3
        gb = DecisionTree(learning_rate=1, max_leaves=2 ** depth, max_depth=depth)
        
        gb.fit(dataset)
        
        clf = DecisionTreeRegressor(max_depth=depth)
        clf = clf.fit(join[x], join[y])
        mse = mean_squared_error(join[y], clf.predict(join[x]))
        self.assertTrue(abs(gb.compute_rmse('test')[0] - math.sqrt(mse)) < 1e-3)
    
    def test_favorita(self):
        con = duckdb.connect(database=':memory:')
        con.execute("CREATE OR REPLACE TABLE holidays AS SELECT * FROM '../data/favorita/holidays.csv';")
        con.execute("CREATE OR REPLACE TABLE oil AS SELECT * FROM '../data/favorita/oil.csv';")
        con.execute("CREATE OR REPLACE TABLE transactions AS SELECT * FROM '../data/favorita/transactions.csv';")
        con.execute("CREATE OR REPLACE TABLE stores AS SELECT * FROM '../data/favorita/stores.csv';")
        con.execute("CREATE OR REPLACE TABLE items AS SELECT * FROM '../data/favorita/items.csv';")
        con.execute("CREATE OR REPLACE TABLE sales AS SELECT * FROM '../data/favorita/sales_small.csv';")
        con.execute("CREATE OR REPLACE TABLE train AS SELECT * FROM '../data/favorita/train_small.csv';")

        y = "Y"
        x = ["htype", "locale", "locale_name", "transferred","f2","dcoilwtico","f3","transactions",
             "f5","city","state","stype","cluster","f4","family","class","perishable","f1"]

        exe = DuckdbExecutor(con, debug=False)
    
        dataset = JoinGraph(exe=exe)
        dataset.add_relation("sales", [], y = 'Y')
        dataset.add_relation("holidays", ["htype", "locale", "locale_name", "transferred","f2"])
        dataset.add_relation("oil", ["dcoilwtico","f3"])
        dataset.add_relation("transactions", ["transactions","f5"])
        dataset.add_relation("stores", ["city","state","stype","cluster","f4"])
        dataset.add_relation("items", ["family","class","perishable","f1"])
        dataset.add_join("sales", "items", ["item_nbr"], ["item_nbr"])
        dataset.add_join("sales", "transactions", ["tid"], ["tid"])
        dataset.add_join("transactions", "stores", ["store_nbr"], ["store_nbr"])
        dataset.add_join("transactions", "holidays", ["date"], ["date"])
        dataset.add_join("holidays", "oil", ["date"], ["date"])

        depth = 3
        reg = DecisionTree(learning_rate=1, max_leaves=2 ** depth, max_depth=depth)

        reg.fit(dataset)

        
        data = pd.read_csv('../data/favorita/train_small.csv')
        clf = DecisionTreeRegressor(max_depth=depth)
        clf = clf.fit(data[x], data[y])
        mse = mean_squared_error(data[y], clf.predict(data[x]))
        math.sqrt(mse)
        self.assertTrue(abs(reg.compute_rmse('train')[0] - math.sqrt(mse)) < 1e-3)
        
    def test_gradient_boosting(self):
        con = duckdb.connect(database=':memory:')
        con.execute("CREATE OR REPLACE TABLE holidays AS SELECT * FROM '../data/favorita/holidays.csv';")
        con.execute("CREATE OR REPLACE TABLE oil AS SELECT * FROM '../data/favorita/oil.csv';")
        con.execute("CREATE OR REPLACE TABLE transactions AS SELECT * FROM '../data/favorita/transactions.csv';")
        con.execute("CREATE OR REPLACE TABLE stores AS SELECT * FROM '../data/favorita/stores.csv';")
        con.execute("CREATE OR REPLACE TABLE items AS SELECT * FROM '../data/favorita/items.csv';")
        con.execute("CREATE OR REPLACE TABLE sales AS SELECT * FROM '../data/favorita/sales_small.csv';")
        con.execute("CREATE OR REPLACE TABLE train AS SELECT * FROM '../data/favorita/train_small.csv';")

        y = "Y"
        x = ["htype", "locale", "locale_name", "transferred","f2","dcoilwtico","f3","transactions",
             "f5","city","state","stype","cluster","f4","family","class","perishable","f1"]

        exe = DuckdbExecutor(con, debug=False)
        depth = 3
        iteration = 3
        dataset = JoinGraph(exe=exe)
        dataset.add_relation("sales", [], y = 'Y')
        dataset.add_relation("holidays", ["htype", "locale", "locale_name", "transferred","f2"])
        dataset.add_relation("oil", ["dcoilwtico","f3"])
        dataset.add_relation("transactions", ["transactions","f5"])
        dataset.add_relation("stores", ["city","state","stype","cluster","f4"])
        dataset.add_relation("items", ["family","class","perishable","f1"])
        dataset.add_join("sales", "items", ["item_nbr"], ["item_nbr"])
        dataset.add_join("sales", "transactions", ["tid"], ["tid"])
        dataset.add_join("transactions", "stores", ["store_nbr"], ["store_nbr"])
        dataset.add_join("transactions", "holidays", ["date"], ["date"])
        dataset.add_join("holidays", "oil", ["date"], ["date"])

        
        reg = GradientBoosting(learning_rate=1, max_leaves=2 ** depth, max_depth=depth, iteration = iteration)

        reg.fit(dataset)

        
        data = pd.read_csv('../data/favorita/train_small.csv')
        clf = GradientBoostingRegressor(max_depth=depth,learning_rate=1, n_estimators=iteration)
        clf = clf.fit(data[x], data[y])
        mse = mean_squared_error(data[y], clf.predict(data[x]))
        math.sqrt(mse)
        self.assertTrue(abs(reg.compute_rmse('train')[0] - math.sqrt(mse)) < 1e-3)

if __name__ == '__main__':
    unittest.main()


from .aggregator import Aggregator
import copy
from .executor import Executor

class JoinGraph:
    def __init__(self, 
                exe, 
                joins = {}, 
                relation_schema = {},
                target_var = None,
                target_relation = None):
        
        self.exe = exe
        # maps each from_relation => to_relation => {keys: (from_keys, to_keys)}
        self.joins = copy.deepcopy(joins)
        # maps each relation => feature => feature_type
        self.relation_schema = copy.deepcopy(relation_schema)
        self.target_var = target_var
        self.target_relation = target_relation
    
    def get_relations(self): 
        return list(self.relation_schema.keys())
    
    def get_relation_schema(self): 
        return self.relation_schema
    
    def get_type(self, relation, feature): 
        return self.relation_schema[relation][feature]

    def get_target_var(self): 
        return self.target_var

    def get_target_relation(self): 
        return self.target_relation
    
    def get_joins(self):
        return self.joins
    
    # add relation, features and target variable to join graph
    # current assumption: Y is in the fact table
    def add_relation(self, relation: str, X: list, y: str = None):
        self.joins[relation] = dict()
        if relation not in self.relation_schema:
                self.relation_schema[relation] = {}
                
        for x in X:
            # by default, assume all features to be numerical
            self.relation_schema[relation][x] = 'NUM'
            
        if y is not None:
            if self.target_var is not None:
                print("Warning: Y already exists and has been replaced")
            self.target_var = y
            self.target_relation = relation
            
    # get features for each table
    def get_relation_features(self, r_name):
        if r_name not in self.relation_schema:
            raise Exception('Attribute not in ' + r_name)
        return list(self.relation_schema[r_name].keys())
    
    # get the join keys between two tables
    # all get all the join keys of one table
    def get_join_keys(self, f_table: str, t_table: str = None):
        if f_table not in self.joins:
            return []
        if t_table:
            if t_table not in self.joins[f_table]:
                raise Exception(t_table, 'not connected to', f_table)
            return self.joins[f_table][t_table]["keys"]
        else:
            keys = set()
            for table in self.joins[f_table]:
                l_keys, _ = self.joins[f_table][table]["keys"]
                keys = keys.union(set(l_keys))
            return list(keys)
    
    # useful attributes are features + join keys
    def get_useful_attributes(self, table):
        useful_attributes = self.get_relation_features(table) + \
                            self.get_join_keys(table)
        return list(set(useful_attributes))
    

    def add_join(self, table_name_left: str, table_name_right: str, left_keys: list, right_keys: list):
        if len(left_keys) != len(right_keys):
            raise Exception('Join keys have different lengths!')
        if table_name_left not in self.relation_schema:
            raise Exception(table_name_left + ' doesn\'t exit!')
        if table_name_right not in self.relation_schema:
            raise Exception(table_name_right + ' doesn\'t exit!')

        left_keys = [attr for attr in left_keys]
        right_keys = [attr for attr in right_keys]

        self.joins[table_name_left][table_name_right] = {"keys": (left_keys, right_keys)}
        self.joins[table_name_right][table_name_left] = {"keys": (right_keys, left_keys)}
        
    def rename(self, table_prev, table_after):
        if table_prev not in self.relation_schema: 
            raise Exception(table_prev + ' doesn\'t exit!')
        if table_after in self.relation_schema: 
            raise Exception(table_after + ' already exits!')
        self.relation_schema[table_after] = self.relation_schema[table_prev]
        del self.relation_schema[table_prev]
        if self.target_relation == table_prev:
            self.target_relation = table_after
            
        for relation in self.joins:
            if table_prev in self.joins[relation]:
                self.joins[relation][table_after] = self.joins[relation][table_prev]
                del self.joins[relation][table_prev]
        
        if table_prev in self.joins:
            self.joins[table_after] = self.joins[table_prev]
            del self.joins[table_prev]
        
    def _preprocess(self):
        self.check_all_features_exist()
    
    def check_all_features_exist(self):
        for table in self.relation_schema:
            features = self.relation_schema[table].keys()
            self.check_features_exist(table, features)
        
    def check_features_exist(self, table, features):
        attributes = self.exe.get_schema(table)
        if not set(features).issubset(set(attributes)):
            Exception('Key error in ' + str(features) + '. Attribute does not exist in table' + table)

#     def decide_feature_type(self, table, attrs, attr_types, threshold, exe: Executor):
#         self.relations.append(table)
#         r_meta = {}
#         for i, attr in enumerate(attrs):
#             if attr_types[i] == 2:
#                 r_meta[attr] = 'NUM'
#             else:
#                 r_meta[attr] = 'CAT'
#                 view = exe.execute_spja_query(aggregate_expressions={attr: (attr, Aggregator.DISTINCT_COUNT)},
#                                               f_table=table)
#                 res = exe.select_all(view)
#                 if res[0][0] <= threshold:
#                     r_meta[attr] = 'LCAT'
#         self.meta_data[table] = r_meta
#         self.r_attrs[table] = list(r_meta.keys())

# TODO: Check fact table and missing join keys
# auto dictionary encoding
# naming could conflict with semi-ring
# for prediction, what if two attributes have the same name?
# semi-join reduction for message to pass to
# todo: remove s,c logic from app


import copy
from .semiring import SemiRing
from .joingraph import JoinGraph
from .executor import Executor
from .aggregator import *


class CJT(JoinGraph):
    def __init__(self, 
                 semi_ring: SemiRing, 
                 join_graph: JoinGraph, 
                 annotations: dict = {}):
        self.message_id = 0
        self.semi_ring = semi_ring
        super().__init__(join_graph.exe,
                         join_graph.joins, 
                         join_graph.relation_schema,
                         join_graph.target_var,
                         join_graph.target_relation)
        # CJT get the join structure from this
        self.annotations = annotations


    def get_message(self, from_table: str, to_table: str):
        return self.joins[from_table][to_table]['message']

    def get_parsed_annotations(self, table):
        if table not in self.annotations:
            return []
        return parse_ann({table: self.annotations[table]})

    def get_all_parsed_annotations(self):
        return parse_ann(self.annotations, True)

    def add_annotations(self, r_name: str, annotation: str):
        if r_name not in self.annotations:
            self.annotations[r_name] = [annotation]
        else:
            self.annotations[r_name].append(annotation)

    def clean_message(self):
        for from_table in self.joins:
            for to_table in self.joins[from_table]:
                if self.joins[from_table][to_table]['message_type'] != Message.IDENTITY:
                    m_name = self.joins[from_table][to_table]['message']
                    self.exe.delete_table(m_name)

    def get_semi_ring(self): 
        return self.semi_ring

    def copy_cjt(self, semi_ring: SemiRing):
        annotations = copy.deepcopy(self.annotations)
        c_cjt = CJT(semi_ring=semi_ring, 
                    join_graph=self, 
                    annotations=annotations)
        return c_cjt

    def get_root_neighbors(self):
        joins, neighbors = self.get_joins(), {}
        for table in joins[self.target_relation]:
            if self.joins[table][self.target_relation]['message_type'] != Message.IDENTITY:
                neighbors[table] = (self.get_join_keys(self.target_relation, table),
                                    self.get_message(table, self.target_relation))
        return neighbors
    
    def calibration(self, rooto_table: str = None):
        if not rooto_table:
            rooto_table = self.target_relation
        self.upward_message_passing(rooto_table, m_type = Message.IDENTITY)
        self.downward_message_passing(rooto_table, m_type = Message.FULL)

    def downward_message_passing(self, 
                                 rooto_table: str = None, 
                                 m_type: Message = Message.UNDECIDED):
        msgs = []
        if not rooto_table:
            rooto_table = self.target_relation
        self._pre_dfs(rooto_table, m_type=m_type)
        return msgs
    
    # TODO: this is not working if upward_message_passing from non-fact table
    def upward_message_passing(self, rooto_table: str = None, 
                               m_type: Message = Message.UNDECIDED):
        if not rooto_table:
            rooto_table = self.target_relation
        self._post_dfs(rooto_table, m_type=m_type)

    def _post_dfs(self, currento_table: str, 
                  parent_table: str = None, 
                  m_type: Message = Message.UNDECIDED):
        jg = self.get_joins()
        if currento_table not in jg:
            return
        for c_neighbor in jg[currento_table]:
            if c_neighbor != parent_table:
                self._post_dfs(c_neighbor, currento_table, m_type=m_type)
        if parent_table:
            self._send_message(from_table=currento_table, to_table=parent_table, m_type=m_type)

    def _pre_dfs(self, currento_table: str, 
                 parent_table: str = None, 
                 m_type: Message = Message.UNDECIDED):
        joins = self.get_joins()
        if currento_table not in joins:
            return
        if currento_table == self.target_relation:
            m_type = Message.FULL
        for c_neighbor in joins[currento_table]:
            if c_neighbor != parent_table:
                self._send_message(currento_table, c_neighbor, m_type=m_type)
                self._pre_dfs(c_neighbor, currento_table, m_type=m_type)

    def absorption(self, table: str, group_by: list, mode=4):
        from_table_attrs = self.get_relation_features(table)
        incoming_messages, join_conds = self._get_income_messages(table)
        
        aggregate_expressions = self.semi_ring.col_sum()
        for attr in group_by:
            aggregate_expressions[attr] = (table + "." + attr, Aggregator.IDENTITY)
        
        return self.exe.execute_spja_query(aggregate_expressions, 
                                           from_tables=[m['message'] for m in incoming_messages]+[table], 
                                           select_conds=join_conds+self.get_parsed_annotations(table),
                                           group_by=[table + '.' + attr for attr in group_by], 
                                           mode=mode)

    
    # get the incoming message from one table to another
    # key function for message passing, Sec 3.3 of CJT paper
    # allow two types of join condition: 1 is for selection, 2 is for semi-join
    def _get_income_messages(self, 
                             table: str, 
                             excluded_table: str = '', 
                             condition=1):
        incoming_messages, join_conds = [], []
        for neighbour_table in self.joins[table]:
            # to_table could be removed for semi-join
            if neighbour_table != excluded_table:
                incoming_message = self.joins[neighbour_table][table]
                if incoming_message['message_type'] == Message.IDENTITY:
                    continue
                    
                # get the join conditions between from_table and incoming_message
                l_join_keys, r_join_keys = self.get_join_keys(neighbour_table, table)
                incoming_messages.append(incoming_message)
                if condition == 1:
                    join_conds += [incoming_message["message"] + "." + l_join_keys[i] + " IS NOT DISTINCT FROM " +
                                   table + "." + r_join_keys[i] for i in range(len(l_join_keys))]
                if condition == 2:
                    join_conds += ['(' + ','.join([table + '.' + key for key in r_join_keys]) + ') in (SELECT (' 
                                   + ','.join([incoming_message["message"] + '.' + key for key in l_join_keys]) 
                                   + ') FROM ' + incoming_message["message"] + ')']
        return incoming_messages, join_conds


    # 3 message types: identity, selection, FULL
    def _send_message(self, from_table: str, to_table: str, m_type: Message = Message.UNDECIDED):
#         print('--Sending Message from', from_table, 'to', to_table, 'm_type is', m_type)
        # identity message optimization
        if m_type == Message.IDENTITY:
            self.joins[from_table][to_table].update({'message_type': m_type,})
            return
        
        if from_table not in self.joins and to_table not in self.joins[from_table]:
            raise Exception('Table', from_table, 'and table', to_table, 'are not connected')
        
        # join with incoming messages
        incoming_messages, join_conds = self._get_income_messages(from_table, to_table)
        
        # assume fact table. Relax it for many-to-many!!
        if m_type == Message.UNDECIDED:
            if from_table == self.target_relation:
                m_type = Message.FULL
            else:
                m_type = Message.SELECTION
                for message_type in [m['message_type'] for m in incoming_messages]:
                    if message_type == Message.FULL:
                        m_type = Message.FULL

        # get the group_by key for this message
        l_join_keys, _ = self.get_join_keys(from_table, to_table)
        
        # compute aggregation
        aggregation = (self.semi_ring.col_sum() if m_type == Message.FULL else {})
        for attr in l_join_keys:
            aggregation[attr] = (from_table + "." + attr, Aggregator.IDENTITY)
            
        message_name = self.exe.execute_spja_query(aggregation, 
                                                    from_tables=[m['message'] for m in incoming_messages]+[from_table], 
                                                    select_conds=join_conds+self.get_parsed_annotations(from_table),
                                                    group_by=[from_table + '.' + attr for attr in l_join_keys], 
                                                    mode=1)

        self.joins[from_table][to_table].update({'message': message_name, 'message_type': m_type})
    
    # by default, lift the target variale
    def lift(self, var = None):
        if var is None:
            var = self.target_var
        lift_exp = self.semi_ring.lift_exp(var)
        # copy the rest attributes
        for attr in self.get_useful_attributes(self.target_relation):
            lift_exp[attr] = (attr, Aggregator.IDENTITY)
        new_fact_name = self.exe.execute_spja_query(lift_exp,
                                                    [self.target_relation],
                                                    mode = 1)
        self.rename(self.target_relation, new_fact_name)



from abc import ABC, abstractmethod
from copy import deepcopy
from .aggregator import Aggregator, Message
from .joingraph import JoinGraph

'''Handle semi ring in DBMS'''
class SemiRing(ABC):
    type: str

    def __init__(self):
        pass
        
    def __add__(self, other):
        pass
    
    def __sub__(self, other):
        pass

    def multiplication(self, semi_ring):
        pass

    def lift_exp(self, attr):
        pass

    def lift_addition(self, attr, constant_):
        pass

    def get_sr_in_select(self, m_type: Message, f_table: str, in_msgs: list, f_table_attrs: list):
        pass


class varSemiRing(SemiRing):
    def __init__(self, s=0, c=0):
        self.r_pair = (s, c)

    def set_semi_ring(self, TS: float, TC: int):
        self.r_pair = (TS, TC)
        
    def __add__(self, other):
        return varSemiRing(self.r_pair[0] + other.r_pair[0], self.r_pair[1] + other.r_pair[1])
        
    def __sub__(self, other):
        return varSemiRing(self.r_pair[0] - other.r_pair[0], self.r_pair[1] - other.r_pair[1])

    def multiplication(self, semi_ring):
        s, c = semi_ring.get_value()
        self.r_pair = (self.r_pair[0] * c + self.r_pair[1] * s, c * self.r_pair[1])

    def lift_exp(self, s = 's', c = '1', s_after = 's', c_after = 'c'):
        return {s_after: (s, Aggregator.IDENTITY), c_after: (c, Aggregator.IDENTITY)}

    def col_sum(self, s = 's', c = 'c', s_after = 's', c_after = 'c'):
        return {s_after: (s, Aggregator.SUM), c_after: (c, Aggregator.SUM)}

    def get_value(self):
        return self.r_pair
    
